"""Custom JARVIS client."""
import re
import traceback
from datetime import datetime

from dis_snek import Snake, listen
from dis_snek.api.events.discord import MessageCreate, MessageDelete, MessageUpdate
from dis_snek.client.utils.misc_utils import find, find_all
from dis_snek.models.discord.channel import DMChannel
from dis_snek.models.discord.embed import EmbedField
from dis_snek.models.discord.message import Message
from dis_snek.models.discord.user import Member
from dis_snek.models.snek.context import Context, InteractionContext
from jarvis_core.db import q
from jarvis_core.db.models import Autopurge, Autoreact, Roleping, Setting, Warning
from jarvis_core.filters import invites
from jarvis_core.util import build_embed
from jarvis_core.util.ansi import RESET, Fore, Format, fmt
from pastypy import AsyncPaste as Paste

DEFAULT_GUILD = 862402786116763668
DEFAULT_ERROR_CHANNEL = 943395824560394250
DEFAULT_SITE = "https://paste.zevs.me"

ERROR_MSG = """
Command Information:
  Name: {invoked_name}
  Args:
{arg_str}

Callback:
  Args:
{callback_args}
  Kwargs:
{callback_kwargs}
"""

KEY_FMT = fmt(Fore.GRAY)
VAL_FMT = fmt(Fore.WHITE)
CMD_FMT = fmt(Fore.GREEN, Format.BOLD)


class Jarvis(Snake):
    @listen()
    async def on_ready(self) -> None:
        """Lepton on_ready override."""
        print("Logged in as {}".format(self.user))  # noqa: T001
        print("Connected to {} guild(s)".format(len(self.guilds)))  # noqa: T001
        print(  # noqa: T001
            "https://discord.com/api/oauth2/authorize?client_id="
            "{}&permissions=8&scope=bot%20applications.commands".format(self.user.id)
        )

    async def on_command_error(
        self, ctx: Context, error: Exception, *args: list, **kwargs: dict
    ) -> None:
        """Lepton on_command_error override."""
        guild = await self.fetch_guild(DEFAULT_GUILD)
        channel = await guild.fetch_channel(DEFAULT_ERROR_CHANNEL)
        error_time = datetime.utcnow().strftime("%d-%m-%Y %H:%M-%S.%f UTC")
        timestamp = int(datetime.now().timestamp())
        timestamp = f"<t:{timestamp}:T>"
        arg_str = (
            "\n".join(f"    {k}: {v}" for k, v in ctx.kwargs.items()) if ctx.kwargs else "    None"
        )
        callback_args = "\n".join(f"    - {i}" for i in args) if args else "    None"
        callback_kwargs = (
            "\n".join(f"    {k}: {v}" for k, v in kwargs.items()) if kwargs else "    None"
        )
        full_message = ERROR_MSG.format(
            error_time=error_time,
            invoked_name=ctx.invoked_name,
            arg_str=arg_str,
            callback_args=callback_args,
            callback_kwargs=callback_kwargs,
        )
        if len(full_message) >= 1900:
            error_message = "  ".join(traceback.format_exception(error))
            full_message += "Exception: |\n  " + error_message
            paste = Paste(content=full_message)
            await paste.save(DEFAULT_SITE)

            await channel.send(
                f"JARVIS encountered an error at {timestamp}. Log too big to send over Discord."
                f"\nPlease see log at {paste.url}"
            )
        else:
            error_message = "".join(traceback.format_exception(error))
            await channel.send(
                f"JARVIS encountered an error at {timestamp}:"
                f"\n```yaml\n{full_message}\n```"
                f"\nException:\n```py\n{error_message}\n```"
            )
        await ctx.send("Whoops! Encountered an error. The error has been logged.", ephemeral=True)
        return await super().on_command_error(ctx, error, *args, **kwargs)

    # Modlog
    async def on_command(self, ctx: InteractionContext) -> None:
        """Lepton on_command override."""
        if not isinstance(ctx.channel, DMChannel) and ctx.invoked_name not in ["pw"]:
            modlog = await Setting.find_one(q(guild=ctx.guild.id, setting="modlog"))
            if modlog:
                channel = await ctx.guild.fetch_channel(modlog.value)
                args = " ".join(f"{KEY_FMT}{k}:{VAL_FMT}{v}{RESET}" for k, v in ctx.kwargs.items())
                fields = [
                    EmbedField(
                        name="Command",
                        value=f"```ansi\n{CMD_FMT}{ctx.invoked_name}{RESET} {args}\n```",
                        inline=False,
                    ),
                ]
                embed = build_embed(
                    title="Command Invoked",
                    description=f"{ctx.author.mention} invoked a command",
                    fields=fields,
                    color="#fc9e3f",
                )
                embed.set_author(name=ctx.author.username, icon_url=ctx.author.display_avatar.url)
                embed.set_footer(
                    text=f"{ctx.author.username}#{ctx.author.discriminator} | {ctx.author.id}"
                )
                await channel.send(embed=embed)

    async def on_message_edit(self, event: MessageUpdate) -> None:
        """Process on_message_edit events."""
        before = event.before
        after = event.after
        if not before.author.bot:
            modlog = await Setting.find_one(q(guild=after.guild.id, setting="modlog"))
            if modlog:
                if before.content == after.content or before.content is None:
                    return
                channel = before.guild.get_channel(modlog.value)
                fields = [
                    EmbedField(
                        "Original Message",
                        before.content if before.content else "N/A",
                        False,
                    ),
                    EmbedField(
                        "New Message",
                        after.content if after.content else "N/A",
                        False,
                    ),
                ]
                embed = build_embed(
                    title="Message Edited",
                    description=f"{before.author.mention} edited a message",
                    fields=fields,
                    color="#fc9e3f",
                    timestamp=after.edited_timestamp,
                    url=after.jump_url,
                )
                embed.set_author(
                    name=before.author.username,
                    icon_url=before.author.display_avatar.url,
                    url=after.jump_url,
                )
                embed.set_footer(
                    text=f"{before.author.username}#{before.author.discriminator} | {before.author.id}"
                )
                await channel.send(embed=embed)
        if not isinstance(after.channel, DMChannel) and not after.author.bot:
            await self.massmention(after)
            await self.roleping(after)
            await self.checks(after)
            await self.roleping(after)
            await self.checks(after)

    async def on_message_delete(self, event: MessageDelete) -> None:
        """Process on_message_delete events."""
        message = event.message
        modlog = await Setting.find_one(q(guild=message.guild.id, setting="modlog"))
        if modlog:
            fields = [EmbedField("Original Message", message.content or "N/A", False)]

            if message.attachments:
                value = "\n".join([f"[{x.filename}]({x.url})" for x in message.attachments])
                fields.append(
                    EmbedField(
                        name="Attachments",
                        value=value,
                        inline=False,
                    )
                )

            if message.sticker_items:
                value = "\n".join([f"Sticker: {x.name}" for x in message.sticker_items])
                fields.append(
                    EmbedField(
                        name="Stickers",
                        value=value,
                        inline=False,
                    )
                )

            if message.embeds:
                value = str(len(message.embeds)) + " embeds"
                fields.append(
                    EmbedField(
                        name="Embeds",
                        value=value,
                        inline=False,
                    )
                )

            channel = message.guild.get_channel(modlog.value)
            embed = build_embed(
                title="Message Deleted",
                description=f"{message.author.mention}'s message was deleted",
                fields=fields,
                color="#fc9e3f",
            )

            embed.set_author(
                name=message.author.username,
                icon_url=message.author.display_avatar.url,
                url=message.jump_url,
            )
            embed.set_footer(
                text=f"{message.author.username}#{message.author.discriminator} | {message.author.id}"
            )
            await channel.send(embed=embed)

    # Events
    async def on_member_join(self, user: Member) -> None:
        """Handle on_member_join event."""
        guild = user.guild
        unverified = await Setting.find_one(q(guild=guild.id, setting="unverified"))
        if unverified:
            role = guild.get_role(unverified.value)
            if role not in user.roles:
                await user.add_roles(role, reason="User just joined and is unverified")

    async def autopurge(self, message: Message) -> None:
        """Handle autopurge events."""
        autopurge = await Autopurge.find_one(q(guild=message.guild.id, channel=message.channel.id))
        if autopurge:
            await message.delete(delay=autopurge.delay)

    async def autoreact(self, message: Message) -> None:
        """Handle autoreact events."""
        autoreact = await Autoreact.find_one(
            q(
                guild=message.guild.id,
                channel=message.channel.id,
            )
        )
        if autoreact:
            for reaction in autoreact.reactions:
                await message.add_reaction(reaction)

    async def checks(self, message: Message) -> None:
        """Other message checks."""
        # #tech
        channel = find(lambda x: x.id == 599068193339736096, message.channel_mentions)
        if channel and message.author.id == 293795462752894976:
            await channel.send(
                content="https://cdn.discordapp.com/attachments/664621130044407838/805218508866453554/tech.gif"  # noqa: E501
            )
        content = re.sub(r"\s+", "", message.content)
        match = invites.search(content)
        setting = await Setting.find_one(q(guild=message.guild.id, setting="noinvite"))
        if not setting:
            setting = Setting(guild=message.guild.id, setting="noinvite", value=True)
            await setting.commit()
        if match:
            guild_invites = await message.guild.invites()
            allowed = [x.code for x in guild_invites] + [
                "dbrand",
                "VtgZntXcnZ",
                "gPfYGbvTCE",
            ]
            if match.group(1) not in allowed and setting.value:
                await message.delete()
                w = Warning(
                    active=True,
                    admin=self.user.id,
                    duration=24,
                    guild=message.guild.id,
                    reason="Sent an invite link",
                    user=message.author.id,
                )
                await w.commit()
                fields = [
                    EmbedField(
                        name="Reason",
                        value="Sent an invite link",
                        inline=False,
                    )
                ]
                embed = build_embed(
                    title="Warning",
                    description=f"{message.author.mention} has been warned",
                    fields=fields,
                )
                embed.set_author(
                    name=message.author.nick if message.author.nick else message.author.name,
                    icon_url=message.author.display_avatar.url,
                )
                embed.set_footer(
                    text=f"{message.author.name}#{message.author.discriminator} | {message.author.id}"  # noqa: E501
                )
                await message.channel.send(embed=embed)

    async def massmention(self, message: Message) -> None:
        """Handle massmention events."""
        massmention = await Setting.find_one(
            q(
                guild=message.guild.id,
                setting="massmention",
            )
        )
        if (
            massmention
            and massmention.value > 0  # noqa: W503
            and len(message.mentions)  # noqa: W503
            - (1 if message.author in message.mentions else 0)  # noqa: W503
            > massmention.value  # noqa: W503
        ):
            w = Warning(
                active=True,
                admin=self.user.id,
                duration=24,
                guild=message.guild.id,
                reason="Mass Mention",
                user=message.author.id,
            )
            await w.commit()
            fields = [EmbedField(name="Reason", value="Mass Mention", inline=False)]
            embed = build_embed(
                title="Warning",
                description=f"{message.author.mention} has been warned",
                fields=fields,
            )
            embed.set_author(
                name=message.author.nick if message.author.nick else message.author.name,
                icon_url=message.author.display_avatar.url,
            )
            embed.set_footer(
                text=f"{message.author.name}#{message.author.discriminator} | {message.author.id}"
            )
            await message.channel.send(embed=embed)

    async def roleping(self, message: Message) -> None:
        """Handle roleping events."""
        rolepings = await Roleping.find(q(guild=message.guild.id, active=True))

        if not rolepings:
            return

        # Get all role IDs involved with message
        roles = []
        for mention in message.role_mentions:
            roles.append(mention.id)
        for mention in message.mentions:
            for role in mention.roles:
                roles.append(role.id)

        if not roles:
            return

        # Get all roles that are rolepinged
        roleping_ids = [r.role for r in rolepings]

        # Get roles in rolepings
        role_in_rolepings = find_all(lambda x: x in roleping_ids, roles)

        # Check if the user has the role, so they are allowed to ping it
        user_missing_role = any(x.id not in roleping_ids for x in message.author.roles)

        # Admins can ping whoever
        user_is_admin = message.author.guild_permissions.ADMINISTRATOR

        # Check if user in a bypass list
        user_has_bypass = False
        for roleping in rolepings:
            if message.author.id in roleping.bypass["users"]:
                user_has_bypass = True
                break
            if any(role.id in roleping.bypass["roles"] for role in message.author.roles):
                user_has_bypass = True
                break

        if role_in_rolepings and user_missing_role and not user_is_admin and not user_has_bypass:
            w = Warning(
                active=True,
                admin=self.user.id,
                duration=24,
                guild=message.guild.id,
                reason="Pinged a blocked role/user with a blocked role",
                user=message.author.id,
            )
            await w.commit()
            fields = [
                EmbedField(
                    name="Reason",
                    value="Pinged a blocked role/user with a blocked role",
                    inline=False,
                )
            ]
            embed = build_embed(
                title="Warning",
                description=f"{message.author.mention} has been warned",
                fields=fields,
            )
            embed.set_author(
                name=message.author.nick if message.author.nick else message.author.name,
                icon_url=message.author.display_avatar.url,
            )
            embed.set_footer(
                text=f"{message.author.name}#{message.author.discriminator} | {message.author.id}"
            )
            await message.channel.send(embed=embed)

    async def on_message(self, event: MessageCreate) -> None:
        """Handle on_message event. Calls other event handlers."""
        message = event.message
        if not isinstance(message.channel, DMChannel) and not message.author.bot:
            await self.autoreact(message)
            await self.massmention(message)
            await self.roleping(message)
            await self.autopurge(message)
            await self.checks(message)
